import io
import tarfile
import warnings
from typing import Sequence, overload
from uuid import UUID

import polars
import requests

from vitalx.aggregation.auth_utils import ExecutorAuth, infer_executor_auth
from vitalx.aggregation.dsl import QueryPartial
from vitalx.aggregation.types import QueryBatch, RelativeTimeframe
from vitalx.types.environments import VitalEnvironmentT, VitalRegionT, api_base_url
from vitalx.types.query import Query, QueryConfig


class Executor:
    environment: VitalEnvironmentT
    region: VitalRegionT
    team_id: UUID
    auth: ExecutorAuth

    def __init__(
        self,
        *,
        environment: VitalEnvironmentT,
        region: VitalRegionT,
        team_id: UUID,
        api_key: str | None = None,
    ) -> None:
        self.auth = infer_executor_auth(team_id=team_id, explicit_api_key=api_key)
        self.team_id = team_id
        self.region = region
        self.environment = environment

        warnings.filterwarnings("ignore", message="Polars found a filename")

    @overload
    def query(
        self,
        timeframe: RelativeTimeframe,
        query: Query | QueryPartial,
        /,
        *,
        user_id: UUID,
        config: QueryConfig = QueryConfig(),
    ) -> tuple[polars.DataFrame]:
        pass

    @overload
    def query(
        self,
        timeframe: RelativeTimeframe,
        query_1: Query | QueryPartial,  # noqa: F841
        query_2: Query | QueryPartial,  # noqa: F841
        /,
        *,
        user_id: UUID,
        config: QueryConfig = QueryConfig(),
    ) -> tuple[polars.DataFrame, polars.DataFrame]:
        pass

    @overload
    def query(
        self,
        timeframe: RelativeTimeframe,
        query_1: Query | QueryPartial,  # noqa: F841
        query_2: Query | QueryPartial,  # noqa: F841
        query_3: Query | QueryPartial,  # noqa: F841
        /,
        *,
        user_id: UUID,
        config: QueryConfig = QueryConfig(),
    ) -> tuple[polars.DataFrame, polars.DataFrame, polars.DataFrame]:
        pass

    @overload
    def query(
        self,
        timeframe: RelativeTimeframe,
        query_1: Query | QueryPartial,  # noqa: F841
        query_2: Query | QueryPartial,  # noqa: F841
        query_3: Query | QueryPartial,  # noqa: F841
        query_4: Query | QueryPartial,  # noqa: F841
        /,
        *,
        user_id: UUID,
        config: QueryConfig = QueryConfig(),
    ) -> tuple[polars.DataFrame, polars.DataFrame, polars.DataFrame, polars.DataFrame]:
        pass

    @overload
    def query(
        self,
        timeframe: RelativeTimeframe,
        query_1: Query | QueryPartial,  # noqa: F841
        query_2: Query | QueryPartial,  # noqa: F841
        query_3: Query | QueryPartial,  # noqa: F841
        query_4: Query | QueryPartial,  # noqa: F841
        query_5: Query | QueryPartial,  # noqa: F841
        /,
        *,
        user_id: UUID,
        config: QueryConfig = QueryConfig(),
    ) -> tuple[
        polars.DataFrame,
        polars.DataFrame,
        polars.DataFrame,
        polars.DataFrame,
        polars.DataFrame,
    ]:
        pass

    @overload
    def query(
        self,
        timeframe: RelativeTimeframe,
        query_1: Query | QueryPartial,  # noqa: F841
        query_2: Query | QueryPartial,  # noqa: F841
        query_3: Query | QueryPartial,  # noqa: F841
        query_4: Query | QueryPartial,  # noqa: F841
        query_5: Query | QueryPartial,  # noqa: F841
        query_6: Query | QueryPartial,  # noqa: F841
        /,
        *,
        user_id: UUID,
        config: QueryConfig = QueryConfig(),
    ) -> tuple[
        polars.DataFrame,
        polars.DataFrame,
        polars.DataFrame,
        polars.DataFrame,
        polars.DataFrame,
        polars.DataFrame,
    ]:
        pass

    @overload
    def query(
        self,
        timeframe: RelativeTimeframe,
        /,
        *queries: Query | QueryPartial,  # noqa: F841
        user_id: UUID,
        config: QueryConfig = QueryConfig(),
    ) -> tuple[polars.DataFrame, ...]:
        pass

    def query(
        self,
        timeframe: RelativeTimeframe,
        /,
        *queries: Query | QueryPartial,
        user_id: UUID,
        config: QueryConfig = QueryConfig(),
    ) -> Sequence[polars.DataFrame]:
        batch = QueryBatch(
            timeframe=timeframe,
            queries=list(
                query if isinstance(query, Query) else query.finalize()
                for query in queries
            ),
            config=config,
        )

        resp = requests.post(
            "{}aggregate/v1/user/{}/query".format(
                api_base_url(self.environment, self.region),
                str(user_id),
            ),
            headers={
                **(self.auth.headers()),
                "Accept": "application/vnd.vital.tar+gzip+parquet",
            },
            json=batch.model_dump(mode="json"),
        )

        resp.raise_for_status()

        df = list[polars.DataFrame]()

        with tarfile.open(fileobj=io.BytesIO(resp.content), mode="r:gz") as tar:
            members = sorted(
                tar.getmembers(), key=lambda m: int(m.name.removesuffix(".parquet"))
            )

            for member in members:
                file = tar.extractfile(member)
                assert file is not None
                df.append(polars.read_parquet(file))

        return df

    def get_result_table(
        self, query_id_or_slug: UUID | str, /, user_id: UUID
    ) -> polars.DataFrame:
        resp = requests.get(
            "{}aggregate/v1/user/{}/continuous_query/{}/result_table".format(
                api_base_url(self.environment, self.region),
                str(user_id),
                str(query_id_or_slug),
            ),
            headers={
                **(self.auth.headers()),
                "Accept": "application/vnd.apache.parquet",
            },
        )

        resp.raise_for_status()

        return polars.read_parquet(resp.content)



from datetime import date
from typing import Any, Literal

from pydantic import BaseModel, Field, model_validator

from vitalx.types.query import Period, Placeholder, Query, QueryConfig


class RelativeTimeframe(BaseModel):
    type: Literal["relative"]
    anchor: date
    past: Period


Timeframe = RelativeTimeframe | Placeholder


class QueryBatch(BaseModel):
    timeframe: Timeframe
    queries: list[Query]
    config: QueryConfig = Field(default_factory=lambda: QueryConfig())

    @model_validator(mode="before")
    @classmethod
    def validate_queries(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values

        # compatibility with SDK 0.1.x or earlier
        if "queries" not in values and (queries := values.pop("instructions", None)):
            values["queries"] = queries

        return values



import polars
import polars._typing

from vitalx.types.query import (
    ActivityColumnT,
    BodyColumnT,
    MealColumnT,
    SleepColumnT,
    TimeseriesFieldsT,
    WorkoutColumnT,
)

DF_GROUP_KEY = "group_key"

SLEEP_DATAFRAME_SCHEMA: dict[SleepColumnT, polars._typing.PolarsDataType] = {
    "session_start": polars.Datetime(time_zone=None, time_unit="ms"),
    "session_end": polars.Datetime(time_zone=None, time_unit="ms"),
    "state": polars.Utf8,
    "duration_second": polars.Int64,
    "stage_asleep_second": polars.Int64,
    "stage_awake_second": polars.Int64,
    "stage_light_second": polars.Int64,
    "stage_rem_second": polars.Int64,
    "stage_deep_second": polars.Int64,
    "stage_unknown_second": polars.Int64,
    "latency_second": polars.Int64,
    "heart_rate_minimum": polars.Int64,
    "heart_rate_mean": polars.Int64,
    "heart_rate_maximum": polars.Int64,
    "heart_rate_resting": polars.Int64,
    "heart_rate_dip": polars.Float64,
    "efficiency": polars.Float64,
    "hrv_mean_rmssd": polars.Float64,
    "hrv_mean_sdnn": polars.Float64,
    "skin_temperature_delta": polars.Float64,
    "respiratory_rate": polars.Float64,
    "score": polars.Int64,
    "source_type": polars.Utf8,
    "source_provider": polars.Utf8,
    "source_app_id": polars.Utf8,
}

ACTIVITY_DATAFRAME_SCHEMA: dict[ActivityColumnT, polars._typing.PolarsDataType] = {
    "date": polars.Date(),
    "calories_total": polars.Float64,
    "calories_active": polars.Float64,
    "steps": polars.Int64,
    "distance_meter": polars.Float64,
    "floors_climbed": polars.Int64,
    "duration_active_second": polars.Int64,
    "intensity_sedentary_second": polars.Int64,
    "intensity_low_second": polars.Int64,
    "intensity_medium_second": polars.Int64,
    "intensity_high_second": polars.Int64,
    "heart_rate_mean": polars.Float64,
    "heart_rate_minimum": polars.Float64,
    "heart_rate_maximum": polars.Float64,
    "heart_rate_resting": polars.Float64,
    "heart_rate_mean_walking": polars.Float64,
    "wheelchair_use": polars.Boolean,
    "wheelchair_push": polars.Int64,
    "source_type": polars.Utf8,
    "source_provider": polars.Utf8,
    "source_app_id": polars.Utf8,
}

WORKOUT_DATAFRAME_SCHEMA: dict[WorkoutColumnT, polars._typing.PolarsDataType] = {
    "session_start": polars.Datetime(time_zone=None, time_unit="ms"),
    "session_end": polars.Datetime(time_zone=None, time_unit="ms"),
    "title": polars.Utf8,
    "sport_name": polars.Utf8,
    "sport_slug": polars.Utf8,
    "duration_active_second": polars.Int64,
    "heart_rate_mean": polars.Int64,
    "heart_rate_minimum": polars.Int64,
    "heart_rate_maximum": polars.Int64,
    "heart_rate_zone_1": polars.Float64,
    "heart_rate_zone_2": polars.Float64,
    "heart_rate_zone_3": polars.Float64,
    "heart_rate_zone_4": polars.Float64,
    "heart_rate_zone_5": polars.Float64,
    "heart_rate_zone_6": polars.Float64,
    "distance_meter": polars.Float64,
    "calories": polars.Float64,
    "elevation_gain_meter": polars.Float64,
    "elevation_maximum_meter": polars.Float64,
    "elevation_minimum_meter": polars.Float64,
    "speed_mean": polars.Float64,
    "speed_maximum": polars.Float64,
    "power_source": polars.Utf8,
    "power_mean": polars.Float64,
    "power_maximum": polars.Float64,
    "power_weighted_mean": polars.Float64,
    "steps": polars.Int64,
    "map_polyline": polars.Utf8,
    "map_summary_polyline": polars.Utf8,
    "source_type": polars.Utf8,
    "source_provider": polars.Utf8,
    "source_app_id": polars.Utf8,
}


BODY_DATAFRAME_SCHEMA: dict[BodyColumnT, polars._typing.PolarsDataType] = {
    "measured_at": polars.Datetime(time_zone=None, time_unit="ms"),
    "weight_kilogram": polars.Float64,
    "fat_mass_percentage": polars.Float64,
    "water_percentage": polars.Float64,
    "muscle_mass_percentage": polars.Float64,
    "visceral_fat_index": polars.Float64,
    "bone_mass_percentage": polars.Float64,
    "lean_body_mass_kilogram": polars.Float64,
    "body_mass_index": polars.Float64,
    "waist_circumference_centimeter": polars.Float64,
    "source_type": polars.Utf8,
    "source_provider": polars.Utf8,
    "source_app_id": polars.Utf8,
}

MEAL_DATAFRAME_SCHEMA: dict[MealColumnT, polars._typing.PolarsDataType] = {
    # Common
    "date": polars.Date,
    "name": polars.Utf8,
    "source_type": polars.Utf8,
    "source_provider": polars.Utf8,
    "source_app_id": polars.Utf8,
    # Nutritional data
    "calories": polars.Float64,
    # Macros
    "carbohydrate_gram": polars.Float64,
    "protein_gram": polars.Float64,
    "alcohol_gram": polars.Float64,
    "fibre_gram": polars.Float64,
    "sugar_gram": polars.Float64,
    "cholesterol_gram": polars.Float64,
    # Fats
    "saturated_fat_gram": polars.Float64,
    "monounsaturated_fat_gram": polars.Float64,
    "polyunsaturated_fat_gram": polars.Float64,
    "omega3_fat_gram": polars.Float64,
    "omega6_fat_gram": polars.Float64,
    "total_fat_gram": polars.Float64,
    # Minerals
    "sodium_milligram": polars.Float64,
    "potassium_milligram": polars.Float64,
    "calcium_milligram": polars.Float64,
    "phosphorus_milligram": polars.Float64,
    "magnesium_milligram": polars.Float64,
    "iron_milligram": polars.Float64,
    "zinc_milligram": polars.Float64,
    "fluoride_milligram": polars.Float64,
    "chloride_milligram": polars.Float64,
    # Vitamins
    "vitamin_a_milligram": polars.Float64,
    "vitamin_b1_milligram": polars.Float64,
    "riboflavin_milligram": polars.Float64,
    "niacin_milligram": polars.Float64,
    "pantothenic_acid_milligram": polars.Float64,
    "vitamin_b6_milligram": polars.Float64,
    "biotin_microgram": polars.Float64,
    "vitamin_b12_microgram": polars.Float64,
    "vitamin_c_milligram": polars.Float64,
    "vitamin_d_microgram": polars.Float64,
    "vitamin_e_milligram": polars.Float64,
    "vitamin_k_microgram": polars.Float64,
    "folic_acid_microgram": polars.Float64,
    # Trace Elements
    "chromium_microgram": polars.Float64,
    "copper_milligram": polars.Float64,
    "iodine_microgram": polars.Float64,
    "manganese_milligram": polars.Float64,
    "molybdenum_microgram": polars.Float64,
    "selenium_microgram": polars.Float64,
}


TIMESERIES_DATAFRAME_SCHEMA: dict[TimeseriesFieldsT, polars._typing.PolarsDataType] = {
    # Common columns
    "source_provider": polars.Utf8,
    "source_type": polars.Utf8,
    "source_workout_id": polars.Utf8,
    "source_sport": polars.Utf8,
    "timezone_offset": polars.Int64,
    # Common but as fields
    "value": polars.Float64,
    "type": polars.Utf8,
    # Subtype specific fields
    # Blood Pressure
    "diastolic": polars.Float64,
    "systolic": polars.Float64,
    # Temperature,
    "sensor_location": polars.Utf8,
    # Workout duration
    "intensity": polars.Utf8,
    # Note
    "tags": polars.List(polars.Utf8),
    "content": polars.Utf8,
}



import typing
from typing import Any, Literal, NamedTuple, cast, get_args, overload

from vitalx.types.query import (
    ActivityColumnExpr,
    ActivityColumnT,
    AggregateExpr,
    BloodPressureTimeseriesExpr,
    BloodPressureTimeseriesFieldT,
    BloodPressureTimeseriesT,
    BodyColumnExpr,
    BodyColumnT,
    ChronotypeValueMacroExpr,
    ColumnExpr,
    DatePartExpr,
    DatePartT,
    DateTimeUnit,
    DateTruncExpr,
    DiscreteTimeseriesExpr,
    DiscreteTimeseriesFieldT,
    DiscreteTimeseriesT,
    GroupByExpr,
    GroupKeyColumnExpr,
    IndexColumnExpr,
    IntervalTimeseriesExpr,
    IntervalTimeseriesFieldT,
    IntervalTimeseriesT,
    NoteTimeseriesExpr,
    NoteTimeseriesFieldT,
    NoteTimeseriesT,
    Period,
    Query,
    SelectExpr,
    SleepColumnExpr,
    SleepColumnT,
    SleepScoreValueMacroExpr,
    SourceColumnExpr,
    SourceFieldT,
    TemperatureTimeseriesExpr,
    TemperatureTimeseriesFieldT,
    TemperatureTimeseriesT,
    TimeseriesExpr,
    TimeseriesT,
    WorkoutColumnExpr,
    WorkoutColumnT,
    WorkoutDurationTimeseriesExpr,
    WorkoutDurationTimeseriesFieldT,
    WorkoutDurationTimeseriesT,
)


def period(value: int, unit: DateTimeUnit) -> Period:
    return Period(value=value, unit=unit)


class SupportsAggregate:
    @property
    def _column_expr(self) -> ColumnExpr:
        return cast(ColumnExpr, self)

    def sum(self) -> AggregateExpr:
        return AggregateExpr(arg=self._column_expr, func="sum")

    def mean(self) -> AggregateExpr:
        return AggregateExpr(arg=self._column_expr, func="mean")

    def min(self) -> AggregateExpr:
        return AggregateExpr(arg=self._column_expr, func="min")

    def max(self) -> AggregateExpr:
        return AggregateExpr(arg=self._column_expr, func="max")

    def median(self) -> AggregateExpr:
        return AggregateExpr(arg=self._column_expr, func="median")

    def stddev(self) -> AggregateExpr:
        return AggregateExpr(arg=self._column_expr, func="stddev")

    def newest(self) -> AggregateExpr:
        return AggregateExpr(arg=self._column_expr, func="newest")

    def oldest(self) -> AggregateExpr:
        return AggregateExpr(arg=self._column_expr, func="oldest")

    def count(self) -> AggregateExpr:
        return AggregateExpr(arg=self._column_expr, func="count")


class IndexColumnExprLike(IndexColumnExpr, SupportsAggregate):
    pass


class SleepColumnExprLike(SleepColumnExpr, SupportsAggregate):
    pass


class ActivityColumnExprLike(ActivityColumnExpr, SupportsAggregate):
    pass


class WorkoutColumnExprLike(WorkoutColumnExpr, SupportsAggregate):
    pass


class BodyColumnExprLike(BodyColumnExpr, SupportsAggregate):
    pass


class SleepScoreValueMacroExprLike(SleepScoreValueMacroExpr, SupportsAggregate):
    pass


class ChronotypeValueMacroExprLike(ChronotypeValueMacroExpr, SupportsAggregate):
    pass


class DiscreteTimeseriesColumnExprLike(DiscreteTimeseriesExpr, SupportsAggregate):
    pass


class IntervalTimeseriesExprLike(IntervalTimeseriesExpr, SupportsAggregate):
    pass


class BloodPressureTimeseriesExprLike(BloodPressureTimeseriesExpr, SupportsAggregate):
    pass


class TemperatureTimeseriesExprLike(TemperatureTimeseriesExpr, SupportsAggregate):
    pass


class WorkoutDurationExprLike(WorkoutDurationTimeseriesExpr, SupportsAggregate):
    pass


class NoteTimeseriesExprLike(NoteTimeseriesExpr, SupportsAggregate):
    pass


class SourceColumnExprLike(SourceColumnExpr, SupportsAggregate):
    pass


TimeseriesExprVar = typing.TypeVar(
    "TimeseriesExprVar",
    bound=TimeseriesExpr,
)


class DiscreteTimeseriesPartial:
    def __init__(self, resource: DiscreteTimeseriesT) -> None:
        self.resource: DiscreteTimeseriesT = resource

    def field(self, name: DiscreteTimeseriesFieldT) -> DiscreteTimeseriesColumnExprLike:
        return DiscreteTimeseriesColumnExprLike(
            field=name,
            timeseries=self.resource,
        )


class IntervalTimeseriesPartial:
    def __init__(self, resource: IntervalTimeseriesT) -> None:
        self.resource: IntervalTimeseriesT = resource

    def field(self, name: IntervalTimeseriesFieldT) -> IntervalTimeseriesExprLike:
        return IntervalTimeseriesExprLike(
            field=name,
            timeseries=self.resource,
        )


class BloodPressureTimeseriesPartial:
    def __init__(self, resource: BloodPressureTimeseriesT) -> None:
        self.resource: BloodPressureTimeseriesT = resource

    def field(
        self, name: BloodPressureTimeseriesFieldT
    ) -> BloodPressureTimeseriesExprLike:
        return BloodPressureTimeseriesExprLike(
            field=name,
            timeseries=self.resource,
        )


class TemperatureTimeseriesPartial:
    def __init__(self, resource: TemperatureTimeseriesT) -> None:
        self.resource: TemperatureTimeseriesT = resource

    def field(self, name: TemperatureTimeseriesFieldT) -> TemperatureTimeseriesExprLike:
        return TemperatureTimeseriesExprLike(
            field=name,
            timeseries=self.resource,
        )


class WorkoutDurationTimeseriesPartial:
    def __init__(self, resource: WorkoutDurationTimeseriesT) -> None:
        self.resource: WorkoutDurationTimeseriesT = resource

    def field(self, name: WorkoutDurationTimeseriesFieldT) -> WorkoutDurationExprLike:
        return WorkoutDurationExprLike(
            field=name,
            timeseries=self.resource,
        )


class NoteTimeseriesPartial:
    def __init__(self, resource: NoteTimeseriesT) -> None:
        self.resource: NoteTimeseriesT = resource

    def field(self, name: NoteTimeseriesFieldT) -> NoteTimeseriesExprLike:
        return NoteTimeseriesExprLike(
            field=name,
            timeseries=self.resource,
        )


class Timeseries:
    @typing.overload
    @staticmethod
    def col(resource: DiscreteTimeseriesT, /) -> DiscreteTimeseriesPartial:
        ...

    @typing.overload
    @staticmethod
    def col(resource: BloodPressureTimeseriesT, /) -> BloodPressureTimeseriesPartial:
        ...

    @typing.overload
    @staticmethod
    def col(resource: TemperatureTimeseriesT, /) -> TemperatureTimeseriesPartial:
        ...

    @typing.overload
    @staticmethod
    def col(
        resource: WorkoutDurationTimeseriesT, /
    ) -> WorkoutDurationTimeseriesPartial:
        ...

    @typing.overload
    @staticmethod
    def col(resource: NoteTimeseriesT, /) -> NoteTimeseriesPartial:
        ...

    @typing.overload
    @staticmethod
    def col(resource: IntervalTimeseriesT, /) -> IntervalTimeseriesPartial:
        ...

    @staticmethod
    def col(resource: TimeseriesT, /) -> Any:
        if resource in get_args(DiscreteTimeseriesT):
            return DiscreteTimeseriesPartial(
                resource=cast(DiscreteTimeseriesT, resource)
            )

        if resource in get_args(IntervalTimeseriesT):
            return IntervalTimeseriesPartial(
                resource=cast(IntervalTimeseriesT, resource)
            )

        if resource in get_args(BloodPressureTimeseriesT):
            return BloodPressureTimeseriesPartial(
                resource=cast(BloodPressureTimeseriesT, resource)
            )

        if resource in get_args(TemperatureTimeseriesT):
            return TemperatureTimeseriesPartial(
                resource=cast(TemperatureTimeseriesT, resource)
            )

        if resource in get_args(WorkoutDurationTimeseriesT):
            return WorkoutDurationTimeseriesPartial(
                resource=cast(WorkoutDurationTimeseriesT, resource)
            )

        if resource in get_args(NoteTimeseriesT):
            return NoteTimeseriesPartial(resource=cast(NoteTimeseriesT, resource))

        raise ValueError(f"Unsupported timeseries type: {resource}.")

    @staticmethod
    def index() -> IndexColumnExprLike:
        """
        The Timeseries datetime index (YYYY-mm-dd HH:mm:ss).
        """
        return IndexColumnExprLike(index="timeseries")


class Source:
    @staticmethod
    def col(name: SourceFieldT) -> SourceColumnExprLike:
        return SourceColumnExprLike(source=name)


class Sleep:
    @staticmethod
    def col(name: SleepColumnT) -> SleepColumnExprLike:
        return SleepColumnExprLike(sleep=name)

    @staticmethod
    def index() -> IndexColumnExprLike:
        """
        The Sleep Session End datetime index (YYYY-mm-dd HH:mm:ss).
        """
        return IndexColumnExprLike(index="sleep")

    @staticmethod
    def score(
        *, version: Literal["automatic"] = "automatic"
    ) -> SleepScoreValueMacroExprLike:
        """
        Computed sleep score using the Vital Horizon AI Sleep Score model.
        """
        return SleepScoreValueMacroExprLike(value_macro="sleep_score", version=version)

    @staticmethod
    def chronotype(
        *, version: Literal["automatic"] = "automatic"
    ) -> ChronotypeValueMacroExprLike:
        """
        Computed chronotype based on the midpoint of the sleep session.
        """
        return ChronotypeValueMacroExprLike(value_macro="chronotype", version=version)


class Activity:
    @staticmethod
    def col(name: ActivityColumnT) -> ActivityColumnExprLike:
        return ActivityColumnExprLike(activity=name)

    @staticmethod
    def index() -> IndexColumnExprLike:
        """
        The Activity calendar date index (YYYY-mm-dd).
        """
        return IndexColumnExprLike(index="activity")


class Workout:
    @staticmethod
    def col(name: WorkoutColumnT) -> WorkoutColumnExprLike:
        return WorkoutColumnExprLike(workout=name)

    @staticmethod
    def index() -> IndexColumnExprLike:
        """
        The Sleep Session End datetime index (YYYY-mm-dd HH:mm:ss).
        """
        return IndexColumnExprLike(index="workout")


class Body:
    @staticmethod
    def col(name: BodyColumnT) -> BodyColumnExprLike:
        return BodyColumnExprLike(body=name)

    @staticmethod
    def index() -> IndexColumnExprLike:
        """
        The Body Measurement datetime index (YYYY-mm-dd HH:mm:ss).
        """
        return IndexColumnExprLike(index="body")


def group_key(offset: int | Literal["*"]) -> GroupKeyColumnExpr:
    """
    The group key columns created by the GroupBy expressions. The offset corresponds
    to the expression declaration order.

    Specify `*` to select all group key columns.
    """
    return GroupKeyColumnExpr(group_key=offset)


@overload
def date_trunc(argument: IndexColumnExpr, period: Period, /) -> DateTruncExpr:
    """
    Truncate the `argument` at the granularity of `period`.

    For example, given these inputs:
    * `argument` is `2024-08-10 12:34:56`
    * `period` is `period(30, "minute")`

    The output is `2024-08-10 12:30:00`.
    """
    ...


@overload
def date_trunc(
    argument: IndexColumnExpr, value: int, unit: DateTimeUnit, /
) -> DateTruncExpr:
    """
    Truncate the `argument` at the granularity of a period length specified by
    `value` and `unit`.

    For example, given these inputs:
    * `argument` is `2024-08-10 12:34:56`
    * Period length is 30 minutes (`value` is 30, and `unit` is `"minute"`)

    The output is `2024-08-10 12:30:00`.
    """
    ...


def date_trunc(argument: IndexColumnExpr, *args: Any) -> DateTruncExpr:
    if len(args) == 1:
        assert isinstance(args[0], Period)
        return DateTruncExpr(arg=argument, date_trunc=args[0])

    if len(args) == 2:
        return DateTruncExpr(
            arg=argument, date_trunc=Period(value=args[0], unit=args[1])
        )

    raise ValueError(f"unsupported input: {args}")


def date_part(argument: IndexColumnExpr, part: DatePartT) -> DatePartExpr:
    """
    Extract a specific date or time component of the datetime `argument`.
    """
    return DatePartExpr(arg=argument, date_part=part)


class QueryPartial(NamedTuple):
    attributes: dict[str, Any]

    def group_by(self, *exprs: GroupByExpr) -> "QueryPartial":
        return QueryPartial({**self.attributes, "group_by": exprs})

    def finalize(self) -> Query:
        return Query.model_validate(self.attributes)


def select(*exprs: SelectExpr) -> QueryPartial:
    return QueryPartial({"select": exprs})



import os
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Protocol, TypedDict
from uuid import UUID

if TYPE_CHECKING:

    class CLIAuthTokens(TypedDict):
        access_token: str

else:
    CLIAuthTokens = dict[str, Any]

try:
    # vitalx.cli_auth is an optional dependency.
    from vitalx.cli_auth import current_tokens

except ModuleNotFoundError:

    def current_tokens(refresh_if_needed: bool = False) -> CLIAuthTokens | None:
        return None


class ExecutorAuth(Protocol):
    def headers(self) -> dict[str, str]:
        ...


@dataclass(frozen=True, eq=True)
class APIKeyAuth(ExecutorAuth):
    api_key: str

    def headers(self) -> dict[str, str]:
        return {"X-Vital-API-Key": self.api_key}


@dataclass(frozen=True, eq=True)
class CLIAuth0Auth(ExecutorAuth):
    team_id: UUID

    def headers(self) -> dict[str, str]:
        tokens = current_tokens(refresh_if_needed=True)
        if not tokens:
            raise RuntimeError("failed to load vitalx-cli auth tokens")

        return {
            "Authorization": "Bearer " + tokens["access_token"],
            "X-Vital-Team-ID": str(self.team_id),
        }


def infer_executor_auth(team_id: UUID, explicit_api_key: str | None = None):
    api_key = explicit_api_key or os.environ.get("VITAL_API_KEY")

    if api_key:
        return APIKeyAuth(api_key=api_key)

    if current_tokens():
        return CLIAuth0Auth(team_id=team_id)

    raise RuntimeError(
        "None of the supported authentication method is available:  Vital API Key, the VITAL_API_KEY env var."
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vitalx', 'vitalx.aggregation']

package_data = \
{'': ['*']}

install_requires = \
['polars>=1.5,<2.0',
 'pydantic>=2.0,<3.0',
 'requests>=2.32,<3.0',
 'vitalx-types<1.0']

extras_require = \
{'cli-auth': ['vitalx-cli-auth<1.0']}

setup_kwargs = {
    'name': 'vitalx-aggregation',
    'version': '0.2.2',
    'description': 'Vital Horizon AI Aggregation API',
    'long_description': 'Vital Horizon AI Aggregation API\n\n* [Product documentation](https://docs.tryvital.io/api-reference/horizon-ai/)\n\n## License\n\nAGPL 3.0. Refer to [the LICENSE](/LICENSE.txt) for more information.\n',
    'author': 'Vital',
    'author_email': 'developers@tryvital.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://www.tryvital.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)
